# coding: utf-8
"""

    ファイル名：   BridgeCurve.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""

import os
import re
import time
import warnings
import numpy                as np
from scipy                  import stats, optimize
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
import seaborn              as sns
from lmfit                  import minimize, Parameters, Parameter, report_fit
from IntensityData          import IntensityData
from GuinierPorodFit        import GuinierPorodFit
from scipy.interpolate      import UnivariateSpline

warnings.filterwarnings("ignore")

DEBUG = False
DEBUG_PLOT = False

SLOPE_WIDTH_HALF = 5
LARGE_VALUE = 1e100
MIN_E_VALUE = 0.1
MIN_F_VALUE = 0.1
MAX_E_VALUE = 50
MAX_F_VALUE = 50
ALT_E_VALUE = 1.5
ALT_F_VALUE = 1.5

def B( a, b, e, f, x ):
    y1 = -b * ( np.power( np.abs( 1 - np.power( np.abs( (a - x)/a ), e ) ), 1/f ) - 1 )
    if e >= ALT_E_VALUE and f >= ALT_F_VALUE:
        return y1

    e_ = max( ALT_E_VALUE, e )
    f_ = max( ALT_F_VALUE, f )

    y2 = -b * ( np.power( np.abs( 1 - np.power( np.abs( (a - x)/a ), e_ ) ), 1/f_ ) - 1 )
    w  = 0.5*np.cos( 2*np.pi*( x/a - 0.5 ) ) + 0.5
    y_ = w*y1 + (1 - w)*y2
    return y_

class BridgeCurve:
    def __init__( self, Q, I, smoother=None ):
        self.Q  = Q
        self.I  = I
        self.smoother = smoother
        # print( 'Q.shape[0]=', Q.shape[0] )
        self.p2_limit = self.Q.shape[0] - SLOPE_WIDTH_HALF*2

    def get_point_index( self, q ):
        return max( 5, min( self.p2_limit, bisect_right( self.Q, q ) ) )

    def estimate_slope_at( self, p ):
        slice_ = slice( p-SLOPE_WIDTH_HALF, p+SLOPE_WIDTH_HALF )
        slope, intercept, r_value, p_value, std_err = stats.linregress( self.Q[slice_], self.I[slice_] )
        return slope

    def get_slice( self, q1, q2 ):
        self.p1 = self.get_point_index( q1 )
        self.p2 = self.get_point_index( q2 )
        self.bridge_slice = slice( self.p1, self.p2 )

        if DEBUG: print( 'p1, p2=', self.p1, self.p2 )
        return slice( self.p1, self.p2 )

    def fit( self, q1, q2, params=None ):

        self.q1 = q1
        self.q2 = q2

        self.get_slice( q1, q2 )
        self.params = params

        points = []
        slopes = []

        if params is None:
            points.append( np.array( [ q1, self.I[self.p1] ] ) )
            points.append( np.array( [ q2, self.I[self.p2] ] ) )
            for p in [ self.p1, self.p2 ]:
                slopes.append( self.estimate_slope_at(p) )
        else:
            G, Rg, D, d = params

            i1 = G * np.exp( -q1**2*Rg**2/3 )
            i2 = D / np.power( q2, d )

            points.append( np.array( [ q1, i1 ] ) )
            points.append( np.array( [ q2, i2 ] ) )

            slope_q1 = i1 * ( -2*q1*Rg**2/3 )
            slope_q2 = - d * i2 / q2
            slopes = [ slope_q1, slope_q2 ]

        origin = self.compute_origin( points, slopes )

        if origin[0] <= points[0][0] or points[1][0] <= origin[0]:
            self.basic_assumption = False
            return

        self.basic_assumption = True

        if DEBUG:
            print( 'slopes=', slopes )
            print( 'points=', points )
            print( 'origin=', origin )

        alpha = 0.3
        qm = (1 - alpha) * q1 + alpha * q2
        im = self.smoother( qm )
        adjust_point = np.array( [qm, im] )

        if DEBUG_PLOT:
            opposite_point = np.zeros(2)
            for i in range(2):
                opposite_point[i] = points[0][i] + points[1][i] - origin[i]
            plt.title( 'Intercection in the Original Linear Plot' )
            plt.plot( self.Q, self.I )
            area_line = []
            for i in range(2):
                area_line.append( [ origin[i], points[1][i], opposite_point[i], points[0][i], origin[i] ] )
            plt.plot( area_line[0], area_line[1], ':', color='black' )
            for point in points:
                plt.plot( point[0], point[1], 'o', color='red' )
            plt.plot( adjust_point[0], adjust_point[1], 'o', color='orange' )
            plt.plot( origin[0], origin[1], 'o', color='yellow' )
            plt.plot( opposite_point[0], opposite_point[1], 'o', color='yellow' )
            plt.show()
            self.opposite_point = opposite_point
            self.area_line      = np.array( area_line )

        self.points = points
        self.slopes = slopes
        self.compute_transform_params( points, origin, adjust_point )

    def compute_origin( self, points, slopes ):
        M = np.array( [ [ slopes[0], -1 ],
                        [ slopes[1], -1 ],
                      ],
                    )

        v = np.array( [ [ slopes[0]*points[0][0] - points[0][1] ],
                        [ slopes[1]*points[1][0] - points[1][1] ],
                      ],
                    )

        Minv = np.linalg.inv( M )
        origin = np.squeeze( np.dot( Minv, v ) )
        return origin

    def compute_transform_params( self, points, origin, adjust_point ):
        b = np.sqrt( ( points[0][0]- origin[0] )**2 + ( points[0][1] - origin[1] )**2 )
        a = np.sqrt( ( points[1][0]- origin[0] )**2 + ( points[1][1] - origin[1] )**2 )

        if DEBUG: print( 'a, b=', a, b )

        """
        u -> v = A * u + o

        A * u = v - o
        u = Ainv * ( v - o )

            u         v - o
        ---------------------
        ( 0, b ).T  P1 - O
        ( a, 0 ).T  P2 - O

        A = ( a1, a2 )

        a1 = ( P2 - O ) / a
        a2 = ( P1 - O ) / b
        """
        a1 = ( points[1] - origin ) / a
        a2 = ( points[0] - origin ) / b

        A = np.array( [ a1, a2 ] ).T
        Ainv = np.linalg.inv( A )

        um = np.squeeze( np.dot( Ainv, (adjust_point - origin).reshape( (2, 1) ) ) )

        # Affine Transformation
        bx  = self.Q[self.bridge_slice]
        by  = self.I[self.bridge_slice]
        bx_ = bx - origin[0]
        by_ = by - origin[1]
        t = np.dot( Ainv, np.vstack( [ bx_, by_ ] ) )
        tx  = t[0,:]
        ty  = t[1,:]
        # print( 'tx=', tx )
        tx_neg = tx < 0
        where = np.where( tx_neg )[0]
        if len( where ) > 0:
            neg_max = where[-1]
            print( 'neg_max=',neg_max )
            pos_start = neg_max + 1
            if pos_start < len(tx):
                tx_max = tx[pos_start]
            else:
                tx_max = a * 1e-3
            txn = np.linspace( 0, tx_max, pos_start + 1 )
            tx[0:pos_start] = txn[0:pos_start]
            # print( 'fixed tx=', tx )
        e, f = self.determine_bridge_exponents( a, b, tx, ty )
        uy  = B( a, b, e, f, tx )

        if DEBUG_PLOT:
            a_area_line = np.dot( Ainv, self.area_line - origin.reshape( (2,1) ) )
            fig = plt.figure()
            ax = fig.add_subplot( 111 )
            ax.set_title( 'Bridge Curve (Superellipse) in the Affine-transformed Plot' )
            ax.plot( a_area_line[0], a_area_line[1], ':', color='black' )
            ax.plot( tx, ty )
            ax.plot( tx, uy, color='pink' )
            for point in [ (0, 0), (a, b) ]:
                ax.plot( point[0], point[1], 'o', color='yellow' )
            for point in [ (0, b), (a, 0) ]:
                ax.plot( point[0], point[1], 'o', color='red' )
            ax.plot( um[0], um[1], 'o', color='orange' )

            xmin, xmax = ax.get_xlim()
            xmin_ = xmax/20
            if -xmin < xmin_:
                ax.set_xlim( -xmin_, xmax )

            an_x = a/10
            an_y = B( a, b, e, f, an_x )

            ymin, ymax = ax.get_ylim()
            xoffset = ( xmax + xmin_ ) / 3
            yoffset = ( ymax - ymin ) / 4

            text = r'$|\frac{x}{%.3g}|^{%.3g}+|\frac{y}{%.3g}|^{%.3g}=1$' % ( a, e, b, f )
            ax.annotate( text, xy=(an_x, an_y),
                            xytext=( an_x + xoffset, an_y + yoffset ),
                            ha='center',
                            arrowprops=dict( headwidth=5, width=0.5, facecolor='black', shrink=0.05),
                            fontsize=16,
                            )

            plt.show()

        bv = np.dot( A, np.vstack( [ tx, uy ] ) ) + origin.reshape( (2,1) )
        self.bridge_smoother = UnivariateSpline( bv[0,:], bv[1,:], s=0 )

        self.a  = a
        self.b  = b
        self.e  = e
        self.f  = f
        self.Ainv = Ainv
        self.A  = A
        self.origin = origin
        self.t  = t

        if DEBUG_PLOT:
            plt.title( 'Bridge Curve in the Original Linear Plot' )
            plt.plot( self.Q, self.I )
            plt.plot( self.area_line[0], self.area_line[1], ':', color='black' )
            by = self.predict( bx )
            plt.plot( bx, by, color='pink' )
            if False:
                plt.plot( bv[0,:], bv[1,:], ':', color='yellow' )
                ys = self.bridge_smoother( bx )
                plt.plot( bx, ys, ':', color='green' )
            for point in points:
                plt.plot( point[0], point[1], 'o', color='red' )
            plt.plot( adjust_point[0], adjust_point[1], 'o', color='orange' )
            for point in [ origin, self.opposite_point ]:
                plt.plot( point[0], point[1], 'o', color='yellow' )
            plt.show()

    def determine_bridge_exponents( self, a, b, tx, ty ):

        def diff_eval( p ):
            e,f = p
            diff = ty - B( a, b, e, f, tx )
            return np.sum( diff**2 )

        e_init = 3
        f_init = 5
        e, f = optimize.fmin( diff_eval, [ e_init, f_init ], disp=False )
        if DEBUG or DEBUG_PLOT: print( 'e, f=', e, f )
        return min(MAX_E_VALUE, max( MIN_E_VALUE, e ) ), min( MAX_F_VALUE, max( MIN_F_VALUE, f ) )

    def predict( self, x ):
        assert( ( self.q1 <= x ).all() and ( x <= self.q2 ).all() )
        if not self.basic_assumption:
            return np.ones( len(x) ) * LARGE_VALUE / ( self.q2 - self.q1 )

        y = self.bridge_smoother( x )
        y_isnan = np.isnan( y )
        if y_isnan.any():
            # TODO: analyze this situation
            print( "WARNINIG: Bridge curve has been modified to remove nan's!" )
            # print( 'y=', y )
            i = np.where( y_isnan )[0][-1] + 1
            if i < len(x):
                mx = x[i]
            else:
                # print( 'y=', y )
                mx = self.q2
            my = self.bridge_smoother( mx )
            y0 = self.points[0][1]
            slope = ( my - y0 ) / ( mx - self.q1 )
            y1 = y0 + slope * ( x[0:i] - self.q1 )
            y = np.hstack( [ y1, y[i:] ] )
            # print( 'fixed y=', y )
        return y

DEMO_PLOT_SMOOTHED_CURVE = False

def demo_plot_bridge( file, q1, q2, auto=False, debug_plot=False ):
    global DEBUG_PLOT

    DEBUG_PLOT = debug_plot

    intensity = IntensityData( file, add_smoother=True )

    I_ = intensity.smoother( intensity.Q )

    d, f = os.path.split( file )

    bridge = BridgeCurve( intensity.Q, intensity.I, smoother=intensity.smoother )
    bridge.fit( q1, q2 )
    x = np.linspace( q1, q2, 100 )
    y = bridge.predict( x )

    fig = plt.figure( figsize=( 15, 9 ) )
    ax0 = fig.add_subplot( 331 )
    ax1 = fig.add_subplot( 332 )
    ax2 = fig.add_subplot( 333 )
    ax3 = fig.add_subplot( 335 )
    ax4 = fig.add_subplot( 336 )
    ax5 = fig.add_subplot( 338 )
    ax6 = fig.add_subplot( 339 )

    texts   = [ 
                ( '(a)  Linear Plot for ',   '       Q',        'Intensity' ),
                ( '(b)  Gunier Plot for ',           'Q**2',     'Ln(I)' ),
                ( '(c)  Porod Plot for ',            'Ln(Q)',    'Ln(I)' ),
                ( '(d)  Zoomed Gunier Plot for ',    'Q**2',     'Ln(I)' ),
                ( '(e)  Zoomed Porod Plot for ',     'Ln(Q)',    'Ln(I)' ),
                ]
    convfs  = [
                    lambda x, y: ( x, y ),
                    lambda x, y: ( x**2, np.log(y) ),
                    lambda x, y: ( np.log(x), np.log(y) ),
                    lambda x, y: ( x**2, np.log(y) ),
                    lambda x, y: ( np.log(x), np.log(y) ),
                ]

    for i, ax in enumerate( [ ax0, ax1, ax2, ax3, ax4 ] ):
        texts_ = texts[i]
        ax.set_title( texts_[0] + f )

        convf   = convfs[i]
        X_, Y_  = convf( intensity.Q, intensity.I )
        x_, y_  = convf( x, y )
        Xs, Ys  = convf( intensity.Q, I_ )

        ax.plot( X_, Y_ )

        if DEMO_PLOT_SMOOTHED_CURVE:
            if i > 0:
                ax.plot( Xs, Ys, color='yellow' )

        ax.set_xlabel( texts_[1] )
        ax.set_ylabel( texts_[2] )

        for p in [bridge.p1, bridge.p2]:
            Xp, Yp  = convf( intensity.Q[p], intensity.I[p] )
            ax.plot( Xp, Yp,'o', color='red'  )

        ax.plot( x_, y_, color='pink' )

    xmin1, xmax1 = ax1.get_xlim()
    ymin1, ymax1 = ax1.get_ylim()
    p1_3 = bridge.p1 + 50
    xmax3, ymin3 = convfs[3]( intensity.Q[p1_3], intensity.I[p1_3] )
    ax3.set_xlim( -0.0001, xmax3 )
    ax3.set_ylim( ymin3, ymax1 )

    xmin2, xmax2 = ax2.get_xlim()
    ymin2, ymax2 = ax2.get_ylim()
    p2_4 = bridge.p2 - 100
    xmin4, ymax4 = convfs[4]( intensity.Q[p2_4], intensity.I[p2_4] )
    ax4.set_xlim( xmin4, xmax2 )
    ax4.set_ylim( ymin2, ymax4 )

    gx, gy = convfs[3]( intensity.Q[0:p1_3], I_[0:p1_3] )
    guinier_region_eval_plot( ax5, ax3.get_xlim(), gx, gy, np.log(q1) )

    px, py = convfs[4]( intensity.Q[p2_4:], I_[p2_4:] )
    porod_region_eval_plot( ax6, ax4.get_xlim(), px, py, np.log(q2) )

    plt.tight_layout()
    plt.show( block=not auto )
    if auto:
        time.sleep( 0.5 )
        plt.close()

def guinier_region_eval_plot( ax, xlim, x, y, q1 ):
    ax.set_title( '(f)  Guinier Region Evaluation' )

    start = 10
    min_width = 10
    r_array = []
    for i in range( len(x) ):
        if i < start + min_width:
            r_value = -1
        else:
            slice_ = slice( start, i )
            slope, intercept, r_value, p_value, std_err = stats.linregress( x[slice_], y[slice_] )

        r_array.append( r_value )

    ax.plot( x, r_array )
    ax.set_xlim( xlim )

def porod_region_eval_plot( ax, xlim, x, y, q2 ):
    ax.set_title( '(g)  Porod Region Evaluation' )

    d_ = None
    end = len(x) - 10
    r_array = []
    for i in range( len(x) ):
        if i > end:
            r_value = -1
        else:
            slice_ = slice( i, None )
            slope, intercept, r_value, p_value, std_err = stats.linregress( x[slice_], y[slice_] )
            if d_ is None and x[i] >= q2:
                d_ = slope

        r_array.append( r_value )

    print( 'Estimated Porod Exponent=', d_ )

    ax.plot( x, r_array )

    ax.set_xlim( xlim )

